/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-lists.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"




/* FIXLIST e n */
/* VARLIST e n */
void reduce_list(NODEPTR e)
{
    NODEPTR loc;
    int     cell_type,
        size_type,
        current_size2,
        current_size1,
        max_size1;
    NODEPTR n = reduce_val (node_lnk_2(e));
    NODEPTR size_loc1;
    NODEPTR list1;
    
    if (Ready) {
        type_check (n, INT);
        loc = reduce_ref (node_lnk_1(e));
        if (Ready) {
            cell_type = is_static_mem (loc) ? STATIC : STATE;
            size_type = (node_type(e) == VARLIST) ? STATE : STATIC;
            if (mem_defined1(loc)) {
                assert0 (is_list (list1 = get_list (loc)),
                         "attempt to overwrite non-list with list");
                current_size2 = node_ival (n);
                size_loc1 = list_size (list1);
                if (!mem_defined1 (size_loc1)) {
                    current_size1 = max_size(list1);
                    if (size_type == STATIC) {
                        assert2d ( current_size1 == current_size2,
                                   "attempt to assign list of fixed size %d to one of max size %d",
                                   current_size2, current_size1);
                    }
                    else {
                        assert2d ( current_size1 == current_size2,
                                   "attempt to assign list of max size %d to one of max size %d",
                                   current_size2, current_size1);
                    }
                }
                else {
                    current_size1 = get_int (size_loc1);
                    max_size1 = max_size(list1);
                    if ( max_size1 != current_size1 ) {
                        if (size_type == STATIC) {
                            assert2d ( current_size1 == current_size2,
                                       "attempt to assign list of fixed size %d to one of fixed size %d",
                                       current_size2, current_size1);
                        }
                        else {
                            assert2d ( max_size1 == current_size2,
                                       "attempt to assign list of max size %d to one of max size %d",
                                       current_size2, max_size1);
                        }
                    }
                    else {
                        if (size_type == STATIC) {
                            assert2d ( current_size1 == current_size2,
                                       "attempt to assign list of fixed size %d to one of fixed size %d",
                                       current_size2, current_size1);
                        }
                        else {
                            assert2d ( current_size1 <= current_size2,
                                       "attempt to assign list of max size %d to one of fixed size %d",
                                       current_size2, current_size1);
                        }
                    }
                }
            }
            
            assign_struct (loc, new_list2 (cell_type, size_type, n, node_mem_history(loc), is_framed_mem(loc)));
            result (True);
        }
    }
}


/******************\
*                  *
*  list operators  *
*                  *
\******************/

void reduce_STRUCT_val(NODEPTR e)
{
    NODEPTR e1;
    if (node_type(e1 = node_lnk_1(e)) == LIST) {
        Ready = T;
    }
    else {
        e1 = reduce_ref_and_fetch (e1);
        if (Ready)
            result (node1 (STRUCT, e1));
    }
}

void reduce_STRUCT_ref(NODEPTR e)
{
    ref_result (reduce_ref (node_lnk_1(e)));
}

/* produce a static list */
/* MAKELIST elist */
void reduce_MAKELIST(NODEPTR e)
{
    NODEPTR elist = node_lnk_1(e);
    NODEPTR item;
    unsigned int size = elist_size (elist);
    unsigned und;
    NODEPTR blk = mem_block2 (STATIC, size + 1, 1, NoFraming);
    NODEPTR size_loc = node_blk_base(blk),
            base_loc =  (size == 0) ? size_loc : size_loc + offset2(1,1),
            loc = curr_loc (base_loc);
    
    Ready = T; und = T;
    put_mem1 (size_loc, int_node ((int)size));
    while ((Ready || (ForceOutput && und) || (!StateProgressFlag)) && (elist != NULL)) {
             /*for (elist = node_lnk_1(e); elist != NULL; elist = node_lnk_2(elist)) {*/
        item = reduce_ref_and_fetch1 (node_lnk_1(elist));
        und = und && reduce_ref_and_fetch_is_undefined(node_lnk_1(elist));
        if (Ready) {
            if ((Ready= val_defined (item))) {
                put_mem1 (loc, copy_val (item));
            }
        }
        elist = node_lnk_2(elist);
        loc += offset2 (1,1);
            /*else
              return*/
     }
     if (!StateProgressFlag)
        Ready=T;
    if (Ready || (ForceOutput && und))
        result (list_node (blk, size_loc, base_loc));
}

/* MAKERANGELIST exp id n */
void reduce_MAKERANGELIST(NODEPTR e)
{
  
  NODEPTR env = node_lnk_2(e);
  NODEPTR expn = node_lnk_1(e);
  NODEPTR n = reduce_val(node_lnk_3(e));
  int nvall, i;

  if (Ready) {
    type_check (n, INT);
    nvall = node_ival(n);

    /*NODEPTR elist = node_lnk_1(e);*/
    NODEPTR item;
    /*unsigned int size = elist_size (elist);*/
    unsigned und;
    NODEPTR blk = mem_block2 (STATIC, nvall + 1, 1, NoFraming);
    NODEPTR size_loc = node_blk_base(blk),
            base_loc =  (nvall == 0) ? size_loc : size_loc + offset2(1,1),
            loc = curr_loc (base_loc);
    
    Ready = T; und = T;
    put_mem1 (size_loc, int_node (nvall));
    i =0;
    while ((Ready || (ForceOutput && und) || (!StateProgressFlag)) && (i<nvall)) {

      
      item = reduce_ref_and_fetch1 (node2(CALL, node2(LAMBDA,env,expn), node2(COMMA,int_node(i),NULL)));
      und = und && reduce_ref_and_fetch_is_undefined(node2(CALL, node2(LAMBDA,env,expn), node2(COMMA,int_node(i),NULL)));
      /*und && reduce_ref_and_fetch_is_undefined(node2(CALL, node2(LAMBDA,env,expn), int_node(nvall) ) );*/
      if (Ready) {
         if ((Ready = val_defined (item))) {
                put_mem1 (loc, copy_val (item));
            }
        }
        /*elist = node_lnk_2(elist);*/
        loc += offset2 (1,1);
        /*nvall -= 1;*/
        i = i+1;
            /*else
              return*/
     }
     if (!StateProgressFlag)
        Ready=T;
     if (Ready || (ForceOutput && und))
        result (list_node (blk, size_loc, base_loc));
  }
}



int reduce_index(NODEPTR e)
{
    e = reduce_val (e);
    int n;
    
    if (Ready) {
        type_check (e, INT);
        n = node_ival(e);
        assert0 (n>=0 , "index variable should be greater than zero");
        return (node_ival(e));
    }
    return (0);
}

/* SUB e1 e2 */
void reduce_SUB(NODEPTR e)
{
    NODEPTR list;
    int     n = reduce_index (node_lnk_2(e));
    if (Ready) {
        list = reduce_ref_and_fetch (node_lnk_1(e));
        if (Ready) {
            switch (node_type(list)) {
                case LIST:
                    assert2d ((0 <= n) && (n < max_size (list)),
                             "subscript should be less than %d: %d",
			     max_size (list), n);
                    ref_result (adj_loc1(list_base (list) + offset2 (n,node_mem_history(list_base (list)))));
                    break;
                case STRING:
                    ref_result (make_ref (substring (list, n, n)));
                    break;
                default:
                    error_break1 ("cannot subscript expression of type %s",
                            print_name (list));
	    }
	}
    }
}

/* SIZE e, |e| */
void reduce_SIZE(NODEPTR e)
{
    NODEPTR list = reduce_ref_and_fetch (node_lnk_1(e));
    if (Ready) {
        switch (node_type(list)) {
            case LIST:
                ref_result (list_size (list));
                break;
            case STRING:
                ref_result (make_ref (int_node (string_size (list))));
                break;
            default:
                error_break1 ("cannot get size of %s", print_name (list));
	}
    }
}


NODEPTR sublist(NODEPTR list, int m, int n)
{
    int     sub_len, current_size;
    NODEPTR blk;
    NODEPTR size_loc, sub_base_loc;

    size_loc = list_size (list);
    current_size = get_int (size_loc);
    assert2d ((m >= 0) && (n <= current_size),
            "sublist [%d..%d] not within list", m, n);
    sub_len = (n > m) ? n - m : 0;
        /*printf("sublist: sub_len = %d\n",sub_len);
    printf("sublist: list_base = %p\n",list_base (list));
    printf("sublist: offset = %d\n",offset2 (m,node_mem_history(size_loc)));*/
    sub_base_loc = (sub_len == 0) ?  size_loc : list_base (list) + offset2 (m,node_mem_history(size_loc));
        /*printf("sublist: sub_base_loc = %p\n",sub_base_loc);*/
    blk = local_mem2 (STATIC,node_mem_history(size_loc),is_framed_mem(size_loc));
    size_loc = is_next (size_loc) ? next_loc (node_blk_base(blk)) :
        curr_loc (node_blk_base(blk));
    put_mem1 (size_loc, int_node (sub_len));
    return (list_node (blk, size_loc, sub_base_loc));
}

/* sublist of e from e1 to e2-1 */
/* SUBLIST e e1 e2 */
void reduce_SUBLIST(NODEPTR e)
{
    NODEPTR list;
    int     m,
            n;
    m = reduce_index (node_lnk_2(e));
    if (Ready) {
        n = reduce_index (node_lnk_3(e));
        if (Ready) {
            list = reduce_ref_and_fetch (node_lnk_1(e));
            if (Ready) {
                switch (node_type(list)) {
                    case LIST:
                        if ((Ready = list_size_defined (list)))
                            ref_result (make_ref (sublist (list, m, n)));
                        break;
                    case STRING:
                        ref_result (make_ref (substring (list, m, n - 1)));
                        break;
                    default:
                        error_break1 ("cannot subscript expression: %s",
                                      print_name (list));
		}
	    }
	}
    }
}
