/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * test of mamdani fuzzy controllers
 */
 
load "fuzzy_logic".

/*
 


 fis_mamdani = fuzzy_Or_i rule_i
 where
 rule_i = antecedent_i fuzzy_and consequent_i

  fuzzification ; fuzzy inference ; defuzzification
  where
  
  fuzzification: 
      in  : membership function,
            list of crisp values
      out : list of fuzzy values

  fuzzy inference:
     in  : list of fuzzy values
     out : list of fuzzy values

  defuzzification
      in : list of crisp values,
           list of fuzzy values
      out: crisp value
*/


/* run */ define test_1() = {
 exists In_food, In_service, Out_tip,
        food_rancid, food_delicious, 
        service_poor, service_good, service_excellent, 
        tip_cheap, tip_average,  tip_generous,
        Mu_food_rancid, Mu_food_delicious,
        Mu_service_poor, Mu_service_good, Mu_service_excellent,
        Mu_tip_cheap, Mu_tip_average, Mu_tip_generous,
        rule_1, rule_2, rule_3, 
        Y, Ys, Tip : {

   /* list of crisp values */
   
   In_food     = lrange_fl($0$, $0.1$, $10$) and
   In_service  = lrange_fl($0$, $0.1$, $10$) and
   Out_tip     = lrange_fl($0$, $0.3$, $30$) and

   /* fuzzy membership functions */
   
   food_rancid       = trapezoid_mf($0$, $1$, $3$, $7$)  and     
   food_delicious    = trapezoid_mf($3$, $5$, $9$, $10$) and
   
   service_poor      = trapezoid_mf($0$, $1$, $2$, $4$)  and
   service_good      = trapezoid_mf($4$, $5$, $6$, $7$)  and
   service_excellent = trapezoid_mf($6$, $7$, $9$, $10$) and
   
   tip_cheap         = gaussian_mf($2$, $10$)        and
   tip_average       = gaussian_mf($2$, $15$)        and
   tip_generous      = gaussian_mf($2$, $20$)        and

   /* fuzzification,
      in  : membership function,
            list of crisp values
      out : list of fuzzy values
    */
   
   Mu_food_rancid       = map(food_rancid, In_food) and
   Mu_food_delicious    = map(food_delicious, In_food) and
   Mu_service_poor      = map(service_poor, In_service) and
   Mu_service_good      = map(service_good, In_service) and
   Mu_service_excellent = map(service_excellent, In_service) and
   Mu_tip_cheap         = map(tip_cheap, Out_tip) and
   Mu_tip_average       = map(tip_average, Out_tip) and
   Mu_tip_generous      = map(tip_generous, Out_tip) and

  /*  fuzzy inference
     in  : list of fuzzy values
     out : list of fuzzy values
     
     rule_i = antecedent_i f_and consequent_i
     fis_mamdani = f_Or_i rule_i
   */
    rule_1 = l_and_min(l_or_max(Mu_service_poor, Mu_food_rancid),
                       Mu_tip_cheap) and
    rule_2 = l_and_min(Mu_service_good, Mu_tip_average) and
    rule_3 = l_and_min(l_or_max(Mu_service_excellent, Mu_food_delicious),
                       Mu_tip_generous) and
                     
   /* fuzzy aggregation
      in  : list of list of fuzzy values
      out : list of fuzzy values 
   */
   
   Ys = [rule_1, rule_2, rule_3] and
   Y = or_max_list( Ys ) and

   /* defuzzification
      in : list of crisp values,
           list of fuzzy values
      out: crisp value
    */
    
   Tip = centroid(Out_tip, Y) and
   output(Tip) and empty
   
 }
}.

/* run */ define test_2() = {
 exists In_food, In_service, Out_tip,
        food_rancid, food_delicious, 
        service_poor, service_good, service_excellent, 
        tip_cheap, tip_average,  tip_generous,
        Mu_food_rancid, Mu_food_delicious,
        Mu_service_poor, Mu_service_good, Mu_service_excellent,
        Mu_tip_cheap, Mu_tip_average, Mu_tip_generous,
        rule_1, rule_2, rule_3, 
        Y, Ys, Tip : {

   /* list of crisp values */
   /* food is a single crisp value */
   In_food     = [$2$ : i< 101 ] and 
   In_service  = lrange_fl($0$, $0.1$, $10$) and
   Out_tip     = lrange_fl($0$, $0.3$, $30$) and

   /* fuzzy membership functions */
   
   food_rancid       = trapezoid_mf($0$, $1$, $3$, $7$)  and     
   food_delicious    = trapezoid_mf($3$, $5$, $10$, $12$) and
   
   service_poor      = trapezoid_mf($0$, $1$, $2$, $4$)  and
   service_good      = trapezoid_mf($4$, $5$, $6$, $7$)  and
   service_excellent = trapezoid_mf($6$, $7$, $9$, $10$) and
   
   tip_cheap         = gaussian_mf($2$, $10$)        and
   tip_average       = gaussian_mf($2$, $15$)        and
   tip_generous      = gaussian_mf($2$, $20$)        and

   /* fuzzification,
      in  : membership function,
            list of crisp values
      out : list of fuzzy values
    */
   
   Mu_food_rancid       = map(food_rancid, In_food) and
   Mu_food_delicious    = map(food_delicious, In_food) and
   Mu_service_poor      = map(service_poor, In_service) and
   Mu_service_good      = map(service_good, In_service) and
   Mu_service_excellent = map(service_excellent, In_service) and
   Mu_tip_cheap         = map(tip_cheap, Out_tip) and
   Mu_tip_average       = map(tip_average, Out_tip) and
   Mu_tip_generous      = map(tip_generous, Out_tip) and

  /*  fuzzy rules
     in  : list of fuzzy values
     out : list of fuzzy values
   */
    rule_1 = l_and_min(l_or_max(Mu_service_poor, Mu_food_rancid),
                       Mu_tip_cheap) and
    rule_2 = l_and_min(Mu_service_good, Mu_tip_average) and
    rule_3 = l_and_min(l_or_max(Mu_service_excellent, Mu_food_delicious),
                       Mu_tip_generous) and
                     
   /* fuzzy aggregation
      in  : list of list of fuzzy values
      out : list of fuzzy values 
   */
   
   Ys = [rule_1, rule_2, rule_3] and
   Y = or_max_list( Ys ) and

   /* defuzzification
      in : list of crisp values,
           list of fuzzy values
      out: crisp value
    */
    
   Tip = centroid(Out_tip, Y) and
   output(Tip) and empty
   
 }
}.

/* run */ define test_3() = {
 exists In_food, In_service, Out_tip,
        food_rancid, food_delicious, 
        service_poor, service_good, service_excellent, 
        tip_cheap, tip_average,  tip_generous,
        Mu_food_rancid, Mu_food_delicious,
        Mu_service_poor, Mu_service_good, Mu_service_excellent,
        Mu_tip_cheap, Mu_tip_average, Mu_tip_generous,
        rule_1, rule_2, rule_3, 
        Y, Ys, Tip : {

   /* list of crisp values */
   /* food and service are single crisp values */
   In_food     = [$6$ : i < 101 ] and 
   In_service  = [$2$ : i < 101 ] and
   Out_tip     = lrange_fl($0$, $0.3$, $30$) and
   
   /* fuzzy membership functions */
   
   food_rancid       = trapezoid_mf($0$, $1$, $3$, $5$)  and     
   food_delicious    = trapezoid_mf($3$, $7$, $10$, $11$) and
   
   service_poor      = trapezoid_mf($0$, $1$, $2$, $4$)  and
   service_good      = trapezoid_mf($4$, $5$, $6$, $7$)  and
   service_excellent = trapezoid_mf($6$, $7$, $9$, $11$) and

   
   
   tip_cheap         = gaussian_mf($2$, $10$)        and
   tip_average       = gaussian_mf($2$, $15$)        and
   tip_generous      = gaussian_mf($2$, $20$)        and

   /* fuzzification,
      in  : membership function,
            list of crisp values
      out : list of fuzzy values
    */
   
   Mu_food_rancid       = map(food_rancid, In_food) and
   Mu_food_delicious    = map(food_delicious, In_food) and
   Mu_service_poor      = map(service_poor, In_service) and
   Mu_service_good      = map(service_good, In_service) and
   Mu_service_excellent = map(service_excellent, In_service) and
   Mu_tip_cheap         = map(tip_cheap, Out_tip) and
   Mu_tip_average       = map(tip_average, Out_tip) and
   Mu_tip_generous      = map(tip_generous, Out_tip) and

  /*  fuzzy rules
     in  : list of fuzzy values
     out : list of fuzzy values
   */
    rule_1 = l_and_min(l_or_max(Mu_service_poor, Mu_food_rancid),
                       Mu_tip_cheap) and
    rule_2 = l_and_min(Mu_service_good, Mu_tip_average) and
    rule_3 = l_and_min(l_or_max(Mu_service_excellent, Mu_food_delicious),
                       Mu_tip_generous) and
                    
   /* fuzzy aggregation
      in  : list of list of fuzzy values
      out : list of fuzzy values 
   */
   
   Ys = [rule_1, rule_2, rule_3] and
   Y = or_max_list( Ys ) and

   
   /* defuzzification
      in  : list of crisp values,
            list of fuzzy values
      out : crisp value
    */
    
   Tip = centroid(Out_tip, Y) and
   output(Tip) and empty
   
 }
}.

set precision = 2.

/* run */ define test_4() = {
 exists in_food, in_service, out_tip,
        food_rancid, food_delicious,
        V_service_poor, V_service_good, V_service_excellent, 
        tip_cheap, tip_average,  tip_generous,
        mu_food_rancid, mu_food_delicious,
        Mu_service_poor, Mu_service_good, Mu_service_excellent,
        mu_tip_cheap, mu_tip_average, mu_tip_generous,
        Antecdent_1, Antecedent_3, 
        Rule_1, Rule_2, Rule_3, 
        Y, Ys, Tip,
        Service_c_poor_1, Service_c_poor_2, 
        Service_c_good_1, Service_c_good_2,
        Service_c_excellent_1,   Service_c_excellent_2,
        C_service_poor, C_service_good, C_service_excellent : {

   /* parameters for service that change over time */

   Service_c_poor_1 = $1$ and Service_c_poor_2 = $2$ and
   Service_c_good_1 = $5$ and Service_c_good_2 = $6$ and
   Service_c_excellent_1 = $8$ and Service_c_excellent_2 = $9$ and

   Service_c_poor_1 gets Service_c_poor_1 - $0.1$ and
   Service_c_poor_2 gets Service_c_poor_2 + $0.1$ and
   Service_c_good_1 gets Service_c_good_1 - $0.1$ and
   Service_c_good_2 gets Service_c_good_2 + $0.1$ and
   Service_c_excellent_1 gets Service_c_excellent_1 - $0.1$ and
   Service_c_excellent_2 gets Service_c_excellent_2 + $0.1$ and

   

   len(9) and

   food_rancid       = trapezoid_mf($0$, $1$, $3$, $4$)  and     
   food_delicious    = trapezoid_mf($3$, $6$, $10$, $11$) and
   tip_cheap         = gaussian_mf($2$, $10$)        and
   tip_average       = gaussian_mf($2$, $17$)        and
   tip_generous      = gaussian_mf($2$, $25$)        and


    /* list of crisp values */
    
    in_food     = lrange_fl($0$, $0.2$, $10$) and
    in_service  = lrange_fl($0$, $0.2$, $10$) and
    out_tip     = lrange_fl($0$, $0.6$, $30$) and

   /* fuzzification,
      in  : membership function,
            list of crisp values
      out : list of fuzzy values
    */
    
   mu_food_rancid       = map(food_rancid, in_food) and
   mu_food_delicious    = map(food_delicious, in_food) and
   mu_tip_cheap         = map(tip_cheap, out_tip) and
   mu_tip_average       = map(tip_average, out_tip) and
   mu_tip_generous      = map(tip_generous, out_tip) and

   keep {
    
   output(Service_c_poor_1, Service_c_poor_2,
          Service_c_good_1, Service_c_good_2,
          Service_c_excellent_1, Service_c_excellent_2 ) and

   
   /* fuzzy membership functions */
   
   V_service_poor      = trapezoid_mf($0$, Service_c_poor_1, Service_c_poor_2, $4$)  and
   V_service_good      = trapezoid_mf($3$, Service_c_good_1, Service_c_good_2, $7$)  and
   V_service_excellent = trapezoid_mf($6$, Service_c_excellent_1, Service_c_excellent_2, $11$) and
   
   
   /* fuzzification,
      in  : membership function,
            list of crisp values
      out : list of fuzzy values
    */
   
   Mu_service_poor      = map(V_service_poor, in_service) and
   Mu_service_good      = map(V_service_good, in_service) and
   Mu_service_excellent = map(V_service_excellent, in_service) and
   
   C_service_poor = centroid(in_service, Mu_service_poor) and 
   output C_service_poor and

   C_service_good = centroid(in_service, Mu_service_good) and 
   output C_service_good and

   C_service_excellent = centroid(in_service, Mu_service_excellent) and 
   output C_service_excellent and
   
  /*  fuzzy inference
     in  : list of fuzzy values
     out : list of fuzzy values
     
     rule_i = antecedent_i f_and consequent_i
     fis_mamdani = f_Or_i rule_i
   */
    Antecedent_1 = l_or_max(Mu_service_poor, mu_food_rancid) and
    Antecedent_3 = l_or_max(Mu_service_excellent, mu_food_delicious) and
    
    Rule_1 = l_and_min(Antecedent_1, mu_tip_cheap) and
    Rule_2 = l_and_min(Mu_service_good, mu_tip_average) and
    Rule_3 = l_and_min(Antecedent_3, mu_tip_generous) and

   /* fuzzy aggregation
      in  : list of list of fuzzy values
      out : list of fuzzy values 
   */
   
   Ys = [Rule_1, Rule_2, Rule_3] 
   and
   Y = or_max_list( Ys ) 

   /* defuzzification
      in : list of crisp values,
           list of fuzzy values
      out: crisp value
    */
    
    and
  
    Tip = centroid(out_tip, Y) and output(Tip)
   
   
  }
   
 }
}.


